#!/usr/bin/env python
"""
_Job_

Data object that describes a job

"""

from builtins import map, range

__all__ = []


from WMCore.DataStructs.Mask import Mask
from WMCore.DataStructs.WMObject import WMObject
from WMCore.Configuration import ConfigSection

import time


class Job(WMObject, dict):
    def __init__(self, name = None, files = None):
        """
        A job has a jobgroup which gives it its subscription and workflow.
        inputFiles is a list containing files associated to a job
        last_update is the time the job last changed
        """
        dict.__init__(self)
        self.baggage = ConfigSection("baggage")
        if files == None:
            self["input_files"] = []
        else:
            self["input_files"] = files

        self["id"] = None
        self["jobgroup"] = None
        self["name"] = name
        self["state"] = 'new'
        self["state_time"] = int(time.time())
        self["outcome"] = "failure"
        self["retry_count"] = 0
        self["location"] = None
        self["mask"] = Mask()
        self["task"] = None
        self["fwjr"] = None
        self["fwjr_path"] = None
        self["workflow"] = None
        self["owner"] = None
        self["estimatedJobTime"] = None
        self["estimatedMemoryUsage"] = None
        self["estimatedDiskUsage"] = None
        return

    #  //
    # // Use property to define getter/setter API for task
    #//  makes job.task and job.tast = "value" possible
    task = property(lambda self: self['task'],
                    lambda self, x: self.__setitem__('task',x))

    def save(self):
        """
        _save_

        Bogus method to make this look more like the WMBS JOB.
        """
        return

    def getFiles(self, type = "list"):
        """
        _getFiles_

        Retrieve information about the input files for the job.  The type
        parameter can be set to one of the following:
          list - A list of File objects will be returned
          set - A set of File objects will be returned
          lfn - A list of LFNs will be returned
          id - A list if File IDs will be returned
        """
        if type == "list":
            return self["input_files"]
        elif type == "set":
            return self.makeset(self["input_files"])
        elif type == "lfn":
            def getLFN(file):
                return file["lfn"]

            lfns = list(map(getLFN, self["input_files"]))
            return lfns
        elif type == "id":
            def getID(file):
                return file["id"]

            ids = list(map(getID, self["input_files"]))
            return ids

    def addFile(self, file):
        """
        _addFile_

        Add a file or list of files to the job's input.
        """
        if isinstance(file, list):
            self["input_files"].extend(file)
        else:
            self["input_files"].append(file)

        return

    def changeState(self, newState):
        """
        _changeState_

        Change the state of the job.
        """
        self["state"] = newState
        self["state_time"] = int(time.time())

        return

    def changeOutcome(self, jobOutcome):
        """
        _changeOutcome_

        Change the final outcome of the job, can be either success or fail.
        """
        self["outcome"] = jobOutcome
        return

    def addResourceEstimates(self, jobTime = None, memory = None, disk = None):
        """
        _addResourceEstimates_

        Add to the current resource estimates, if None then initialize them
        to the given value. Each value can be set independently.
        """
        if not self["estimatedJobTime"]:
            self["estimatedJobTime"] = jobTime
        elif jobTime:
            self["estimatedJobTime"] += jobTime

        if not self["estimatedMemoryUsage"]:
            self["estimatedMemoryUsage"] = memory
        elif memory:
            self["estimatedMemoryUsage"] += memory

        if not self["estimatedDiskUsage"]:
            self["estimatedDiskUsage"] = disk
        elif disk:
            self["estimatedDiskUsage"] += disk

        return

    def getBaggage(self):
        """
        _getBaggage_

        Get a reference to the embedded ConfigSection that is
        used to store per job arguments needed to be propagated at
        submission time

        """
        return self.baggage


    def addBaggageParameter(self, attrName, value):
        """
        _addBaggageParameter_

        Add an attribute as process.pset1.pset2.param = value
        Value should be the appropriate python type

        """
        currentPSet = self.baggage
        paramList = attrName.split(".")
        for i in range(0, len(paramList)):
            param = paramList.pop(0)
            if len(paramList) > 0:
                if not hasattr(currentPSet, param):
                    currentPSet.section_(param)
                currentPSet = getattr(currentPSet, param)
            else:
                setattr(currentPSet, param, value)



"""
_DBFormatter_

A class to define the standardised formatting of database results.
Holds a bunch of helper methods to format input and output of sql
interactions.
"""

from builtins import str, bytes, zip, range

import datetime
import time
import types

from Utils.Utilities import decodeBytesToUnicodeConditional
from WMCore.DataStructs.WMObject import WMObject
from Utils.PythonVersion import PY3


class DBFormatter(WMObject):
    def __init__(self, logger, dbinterface):
        """
        The class holds a connection to the database in self.dbi. This is a
        DBCore.DBInterface object.
        """
        self.logger = logger
        self.dbi = dbinterface

    def truefalse(self, value):
        if value in ('False', 'FALSE', 'n', 'N', 'NO', 'No'):
            value = 0
        return bool(value)

    def convertdatetime(self, t):
        return int(time.mktime(t.timetuple()))

    def timestamp(self):
        """
        generate a timestamp
        """
        t = datetime.datetime.now()
        return self.convertdatetime(t)

    def format(self, result):
        """
        Some standard formatting, put all records into a list.
        Returns a list of lists
        """
        out = []
        for r in result:
            for i in r.fetchall():
                row = []
                for j in i:
                    row.append(j)
                out.append(row)
            r.close()
        return out

    def formatOne(self, result):
        """
        Return one record
        """
        out = []
        for r in result:
            for i in r.fetchone():
                out.append(i)
        return out

    def formatDict(self, result):
        """
        Returns an array of dictionaries representing the results
        """
        dictOut = []
        for r in result:
            descriptions = r.keys
            for i in r.fetchall():
                # WARNING: this can generate errors for some stupid reason
                # in both oracle and mysql.
                entry = {}
                for index in range(0, len(descriptions)):
                    # WARNING: Oracle returns table names in CAP!
                    if isinstance(descriptions[index], (str, bytes)):
                        keyName = decodeBytesToUnicodeConditional(descriptions[index], condition=PY3)
                    else:
                        keyName = descriptions[index]
                    if isinstance(i[index], (str, bytes)):
                        entry[keyName.lower()] = decodeBytesToUnicodeConditional(i[index], condition=PY3)
                    else:
                        entry[keyName.lower()] = i[index]

                dictOut.append(entry)

            r.close()

        return dictOut

    def formatList(self, result):
        """
        Returns a flat array with the results.
        Ideally used for single column queries
        """
        listOut = []
        for r in result:
            descriptions = r.keys
            for i in r.fetchall():
                for index in range(0, len(descriptions)):
                    if isinstance(i[index], (str, bytes)):
                        listOut.append(decodeBytesToUnicodeConditional(i[index], condition=PY3))
                    else:
                        listOut.append(i[index])
            r.close()
        return listOut

    def formatOneDict(self, result):
        """
        Return a dictionary representing the first record
        """
        if (len(result) == 0):
            return {}

        r = result[0]
        description = [str(x).lower() for x in r.keys]
        if len(r.data) < 1:
            return {}

        return dict(list(zip(description, r.fetchone())))

    def formatCursor(self, cursor, size=10):
        """
        Fetch the driver cursor directly.
        Tested only with cx_Oracle.
        Cursor must be already executed.
        Use fetchmany(size = default arraysize = 50)

        """
        if isinstance(cursor.keys, types.MethodType):
            keys = [x.lower() for x in cursor.keys()]  # warning: do not modernize this line.
        else:
            keys = [x.lower() for x in cursor.keys]
        result = []
        while True:
            if not cursor.closed:
                rows = cursor.fetchmany(size=size)
                if not rows:
                    cursor.close()
                    break
                for r in rows:
                    result.append(dict(list(zip(keys, r))))
            else:
                break
        if not cursor.closed:
            cursor.close()
        return result

    def getBinds(self, **kwargs):
        binds = {}
        for i in kwargs:
            binds = self.dbi.buildbinds(self.dbi.makelist(kwargs[i]), i, binds)
        return binds

    def execute(self, conn=None, transaction=False, returnCursor=False):
        """
        A simple select with no binds/arguments is the default
        """
        result = self.dbi.processData(self.sql, self.getBinds(),
                                      conn=conn, transaction=transaction,
                                      returnCursor=returnCursor)
        return self.format(result)

    def executeOne(self, conn=None, transaction=False, returnCursor=False):
        """
        A simple select with no binds/arguments is the default
        """
        result = self.dbi.processData(self.sql, self.getBinds(),
                                      conn=conn, transaction=transaction,
                                      returnCursor=returnCursor)
        return self.formatOne(result)



#!/usr/bin/env python

"""
_Run_

container representing a run, and its constituent lumi sections and event counts
"""

from __future__ import print_function
from future.utils import viewitems, listitems

import hashlib
from builtins import str as newstr, bytes as newbytes
from Utils.PythonVersion import PY3
from Utils.Utilities import encodeUnicodeToBytesConditional, encodeUnicodeToBytes
from WMCore.DataStructs.WMObject import WMObject


class Run(WMObject):
    """
    _Run_

    Run container, is a list of lumi sections with associate event counts

    TODO 
    - use the decorator `from functools import total_ordering` after
      dropping support for python 2.6
    - then, drop __ne__, __le__, __gt__, __ge__
    """

    def __init__(self, runNumber=None, *newLumis):
        WMObject.__init__(self)
        self.run = runNumber
        self.eventsPerLumi = {}
        self.extendLumis(newLumis)

    def __str__(self):
        return "Run%s:%s" % (self.run, self.eventsPerLumi)

    def __eq__(self, rhs):
        """
        Check equality of run numbers and then underlying lumi/event dicts
        """
        if not isinstance(rhs, Run):
            return False
        if self.run != rhs.run:
            return False
        return self.eventsPerLumi == rhs.eventsPerLumi

    def __ne__(self, rhs):
        return not self.__eq__(rhs)

    def __lt__(self, rhs):
        """
        Compare on run # first, then by lumis as a list, then by events in each lumi
        """
        # check run number
        if self.run != rhs.run:
            return self.run < rhs.run
        # if same run number, check list of lumi sections
        if sorted(self.eventsPerLumi.keys()) != sorted(rhs.eventsPerLumi.keys()):
            return sorted(self.eventsPerLumi.keys()) < sorted(rhs.eventsPerLumi.keys())
        # if same list of lumis, check events in each lumi section
        for lumiNumber in sorted(self.eventsPerLumi):
            if self.eventsPerLumi[lumiNumber] == rhs.eventsPerLumi.get(lumiNumber):
                continue
            else: 
                return self.eventsPerLumi[lumiNumber] < rhs.eventsPerLumi.get(lumiNumber)
        # if same run number, same lumi sections list, 
        # same events in each lumi section:
        # then the runs are equal and __lt__ should return false
        return False

    def __le__(self, other):
        return self.__lt__(other) or self.__eq__(other)

    def __gt__(self, other):
        return not self.__le__(other)

    def __ge__(self, other):
        return not self.__lt__(other)

    def extend(self, items):
        """
        Redirect to the function that already does this
        """
        self.extendLumis(items)
        return

    def __add__(self, rhs):
        """
        Combine two runs
        """
        if self.run != rhs.run:
            msg = "Adding together two different runs"
            msg += "Run %s does not equal Run %s" % (self.run, rhs.run)
            raise RuntimeError(msg)

        for lumi, events in viewitems(rhs.eventsPerLumi):
            if lumi not in self.eventsPerLumi or not self.eventsPerLumi[lumi]:  # Either doesn't exist, 0, or None
                self.eventsPerLumi[lumi] = events
            else:
                self.eventsPerLumi[lumi] += events
        return self

    def __iter__(self):
        return self.eventsPerLumi.__iter__()

    def __next__(self):
        """
        __next__ no longer needed
        """
        raise NotImplementedError

    def __len__(self):
        """
        Number of lumis
        """
        return self.eventsPerLumi.__len__()

    def __getitem__(self, key):
        """
        Get the nth lumi from the list (no event count)
        """
        return sorted(self.eventsPerLumi.keys()).__getitem__(key)

    def __setitem__(self, key, lumi):
        """
        Replace the nth lumi from the list (no event count)
        """
        try:
            oldLumi = sorted(self.eventsPerLumi.keys())[key]  # Extract the lumi from the sorted list
            del self.eventsPerLumi[oldLumi]  # Delete it and add the new one
        except IndexError:
            pass
        self.appendLumi(lumi)

    def __delitem__(self, key):
        try:
            oldLumi = sorted(self.eventsPerLumi.keys())[key]  # Extract the lumi from the sorted list
            del self.eventsPerLumi[oldLumi]  # Delete it
        except IndexError:
            pass

    def __hash__(self):
        """
        Calculate the value of the hash

        NOTE: Python2 maxint is:
        > python -c 'import sys; print(sys.maxint)'
        9223372036854775807
        so we cannot use the full range of the hexadecimal hash code because
        it could cause an integer overflow. This is the maximum slice/value we
        can safely use:
        > int(15 * "f", base=16)
        1152921504606846975
        """
        if isinstance(self.run, (newstr, newbytes)):
            value = encodeUnicodeToBytesConditional(self.run, condition=PY3)
        else:
            value = encodeUnicodeToBytesConditional(str(self.run), condition=PY3)
        hashValue = hashlib.sha1(value)
        # Generate immutable sorted list of lumis
        frozenEvents = str(sorted(listitems(self.eventsPerLumi), key=lambda x: x[0]))
        hashValue.update(encodeUnicodeToBytes(frozenEvents))
        return int(hashValue.hexdigest()[:15], 16)

    @property
    def lumis(self):
        """
        Property that makes existing uses of myRun.lumis function by returning a list
        """
        return sorted(self.eventsPerLumi.keys())

    @lumis.setter
    def lumis(self, lumiList):
        """
        Setter to allow for replacement of the lumis with a list or list of tuples
        """
        self.eventsPerLumi = {}  # Remove existing dictionary
        for lumi in lumiList:
            if isinstance(lumi, (list, tuple)):
                self.eventsPerLumi[lumi[0]] = lumi[1]
            else:
                self.eventsPerLumi[lumi] = None

    def extendLumis(self, lumiList):
        """
        Method to replace myRun.lumis.extend() which does not work with the property
        """
        for lumi in lumiList:
            if not isinstance(lumi, (list, tuple)):  # comma separated lumi numbers
                self.eventsPerLumi[lumi] = None
            else:
                if isinstance(lumi, list) and not isinstance(lumi[0], tuple):  # then it's a plain list
                    for l in lumi:
                        self.eventsPerLumi[l] = None
                else:
                    if isinstance(lumi, tuple):  # it's an unpacked list of tuples
                        lumi = [(lumi)]
                    # it's a list/tuple of tuples
                    for tp in lumi:
                        if tp[0] in self.eventsPerLumi and self.eventsPerLumi[tp[0]]:
                            self.eventsPerLumi[tp[0]] += tp[1]  # Already exists, add events
                        else:  # Doesn't exist or is 0 or None
                            self.eventsPerLumi[tp[0]] = tp[1]

    def appendLumi(self, lumi):
        """
        Method to replace myRun.lumis.append() which does not work with the property
        """
        if isinstance(lumi, (list, tuple)) and self.eventsPerLumi[lumi[0]]:  # Already exists, add events
            self.eventsPerLumi[lumi[0]] += lumi[1]
        elif isinstance(lumi, (list, tuple)):  # Doesn't exist or is 0 or None
            self.eventsPerLumi[lumi[0]] = lumi[1]
        else:  # Just given lumis, not events
            if lumi not in self.eventsPerLumi:  # Don't overwrite existing events
                self.eventsPerLumi[lumi] = None

    def getEventsByLumi(self, lumi):
        """
        getter to select event counts by given lumi
        """
        return self.eventsPerLumi.get(lumi)

    def json(self):
        """
        _json_

        Convert to JSON friendly format.  Include some information for the
        thunker so that we can convert back.
        """
        return {"Run": self.run, "Lumis": self.eventsPerLumi,
                "thunker_encoded_json": True, "type": "WMCore.DataStructs.Run.Run"}

    def __to_json__(self, thunker=None):
        """
        __to_json__

        This is the standard way we JSONize other objects.
        Included here so we have a uniform method.
        """
        return self.json()

    def __from_json__(self, jsondata, thunker):
        """
        __from_json__

        Convert JSON data back into a Run object with integer lumi numbers
        """
        self.run = jsondata["Run"]
        self.eventsPerLumi = {}
        for lumi, events in viewitems(jsondata["Lumis"]):
            self.eventsPerLumi[int(lumi)] = events  # Make the keys integers again

        return self
